//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include <iostream>
#include <fstream>

// #include <random>
#include <yaml-cpp/yaml.h>
#include <Eigen/Dense>
#include <ctime>
#include "eigenmvn.h"

#include "pinocchio/parsers/urdf.hpp"
#include "pinocchio/algorithm/joint-configuration.hpp"
#include "pinocchio/algorithm/kinematics.hpp"
#include "pinocchio/algorithm/center-of-mass.hpp"
#include "pinocchio/algorithm/centroidal.hpp"
#include "pinocchio/algorithm/rnea.hpp"
#include "pinocchio/algorithm/crba.hpp"
#include "pinocchio/algorithm/frames.hpp"

// CNPY
#include <cnpy.h>

// WOLF
#include <core/problem/problem.h>
#include <core/ceres_wrapper/solver_ceres.h>
#include <core/math/rotations.h>
#include <core/capture/capture_base.h>
#include <core/capture/capture_pose.h>
#include <core/feature/feature_base.h>
#include <core/factor/factor_pose_3d_with_extrinsics.h>
#include <core/factor/factor_block_absolute.h>
#include <core/processor/processor_odom_3d.h>
#include <core/sensor/sensor_odom_3d.h>
#include "core/tree_manager/tree_manager_sliding_window_dual_rate.h"
#include "core/yaml/parser_yaml.h"

#include <core/sensor/sensor_pose.h>
#include <core/processor/processor_pose.h>

// IMU
#include "imu/sensor/sensor_imu.h"
#include "imu/capture/capture_imu.h"
#include "imu/processor/processor_imu.h"
#include "imu/factor/factor_imu.h"

// BODYDYNAMICS
#include "bodydynamics/internal/config.h"
#include "bodydynamics/sensor/sensor_inertial_kinematics.h"
#include "bodydynamics/sensor/sensor_force_torque.h"
#include "bodydynamics/capture/capture_inertial_kinematics.h"
#include "bodydynamics/capture/capture_force_torque_preint.h"
#include "bodydynamics/capture/capture_leg_odom.h"
#include "bodydynamics/processor/processor_inertial_kinematics.h"
#include "bodydynamics/processor/processor_force_torque_preint.h"
#include "bodydynamics/factor/factor_inertial_kinematics.h"
#include "bodydynamics/factor/factor_force_torque_preint.h"

#include "bodydynamics/sensor/sensor_point_feet_nomove.h"
#include "bodydynamics/processor/processor_point_feet_nomove.h"
#include "bodydynamics/capture/capture_point_feet_nomove.h"
#include "bodydynamics/factor/factor_point_feet_nomove.h"




using namespace Eigen;
using namespace wolf;
using namespace pinocchio;

typedef pinocchio::SE3Tpl<double> SE3;
typedef pinocchio::ForceTpl<double> Force;

void printFactorCosts(FrameBasePtr kf_last){
    int nb_max_fact_ptf = 4;
    int nb = 0;
    for (auto f: kf_last->getFactorList()){
        auto fimu = std::dynamic_pointer_cast<FactorImu>(f);
        if (fimu){
            std::cout << std::setprecision(12) << "  C_IMU " << sqrt(fimu->cost()) << std::endl;
        }
        else{
            auto fptn = std::dynamic_pointer_cast<FactorPointFeetNomove>(f);
            if (fptn and (nb < nb_max_fact_ptf)){
                Vector3d error = fptn->error();
                double cost = fptn->cost();
                std::cout << std::setprecision(12) << "  E_PF" << nb << " " << error.transpose() << std::endl;
                std::cout << std::setprecision(12) << "  C_PF" << nb << " " << sqrt(cost) << std::endl;
                nb++;
            }
        }
    }
}


int main (int argc, char **argv) {
    // retrieve parameters from yaml file
    YAML::Node config = YAML::LoadFile("/home/mfourmy/Documents/Phd_LAAS/wolf/bodydynamics/demos/solo_real_estimation.yaml");
    double dt = config["dt"].as<double>();
    // double min_t = config["min_t"].as<double>();
    double max_t = config["max_t"].as<double>();
    double solve_every_sec = config["solve_every_sec"].as<double>();
    bool solve_end = config["solve_end"].as<bool>();

    // Ceres setup
    int max_num_iterations = config["max_num_iterations"].as<int>();
    double func_tol = config["func_tol"].as<double>();
    bool compute_cov = config["compute_cov"].as<bool>();
    
    // robot
    std::string robot_urdf = config["robot_urdf"].as<std::string>();
    int dimc = config["dimc"].as<int>();
    int nb_feet = config["nb_feet"].as<int>();


    // priors
    double std_prior_p = config["std_prior_p"].as<double>();
    double std_prior_o = config["std_prior_o"].as<double>();
    double std_prior_v = config["std_prior_v"].as<double>();
    double std_abs_bias_pbc = config["std_abs_bias_pbc"].as<double>();
    double std_abs_bias_acc = config["std_abs_bias_acc"].as<double>();
    double std_abs_bias_gyro = config["std_abs_bias_gyro"].as<double>();    
    double std_bp_drift = config["std_bp_drift"].as<double>();
    std::vector<double> bias_imu_prior_vec = config["bias_imu_prior"].as<std::vector<double>>();
    Eigen::Map<Vector6d> bias_imu_prior(bias_imu_prior_vec.data());
    std::vector<double> i_p_im_vec = config["i_p_im"].as<std::vector<double>>();
    Eigen::Map<Vector3d> i_p_im(i_p_im_vec.data());
    std::vector<double> i_q_m_vec = config["i_q_m"].as<std::vector<double>>();
    Eigen::Map<Vector4d> i_qvec_m(i_q_m_vec.data());
    std::vector<double> m_p_mb_vec = config["m_p_mb"].as<std::vector<double>>();
    Eigen::Map<Vector3d> m_p_mb(m_p_mb_vec.data());
    std::vector<double> m_q_b_vec = config["m_q_b"].as<std::vector<double>>();
    Eigen::Map<Vector4d> m_qvec_b(m_q_b_vec.data());


    // kinematics
    double std_foot_nomove = config["std_foot_nomove"].as<double>();
    double std_altitude = config["std_altitude"].as<double>();
    double foot_radius = config["foot_radius"].as<double>();
    std::vector<double> delta_qa_vec = config["delta_qa"].as<std::vector<double>>();
    Eigen::Map<Eigen::Matrix<double,12,1>> delta_qa(delta_qa_vec.data());
    std::vector<double> alpha_qa_vec = config["alpha_qa"].as<std::vector<double>>();
    Eigen::Map<Eigen::Matrix<double,12,1>> alpha_qa(alpha_qa_vec.data());
    std::vector<double> l_p_lg_vec = config["l_p_lg"].as<std::vector<double>>();
    Eigen::Map<Vector3d> l_p_lg(l_p_lg_vec.data());
    double fz_thresh = config["fz_thresh"].as<double>();
    

    // MOCAP
    double std_pose_p = config["std_pose_p"].as<double>();
    double std_pose_o_deg = config["std_pose_o_deg"].as<double>();
    double std_mocap_extr_p = config["std_mocap_extr_p"].as<double>();
    double std_mocap_extr_o_deg = config["std_mocap_extr_o_deg"].as<double>();
    bool unfix_extr_sensor_pose = config["unfix_extr_sensor_pose"].as<bool>();
    bool time_tolerance_mocap = config["time_tolerance_mocap"].as<double>();
    double time_shift_mocap = config["time_shift_mocap"].as<double>();
    

    std::string data_file_path = config["data_file_path"].as<std::string>();
    std::string out_npz_file_path = config["out_npz_file_path"].as<std::string>();

    // MOCAP to IMU transform
    Quaterniond i_q_m(i_qvec_m);
    i_q_m.normalize();
    assert(i_q_m.normalized().isApprox(i_q_m));
    SE3 i_T_m(i_q_m, i_p_im);
    Matrix3d i_R_m =  i_T_m.rotation();

    // IMU to MOCAP transform
    SE3 m_T_i = i_T_m.inverse();
    Vector3d m_p_mi = m_T_i.translation();
    Quaterniond m_q_i = Quaterniond(m_T_i.rotation());

    // MOCAP to BASE transform
    Quaterniond m_q_b(m_qvec_b);
    m_q_b.normalize();
    assert(m_q_b.normalized().isApprox(m_q_b));
    SE3 m_T_b(m_q_b, m_p_mb);
    Matrix3d m_R_b =  m_T_b.rotation();

    // IMU to BASE transform (composition)
    SE3 i_T_b = i_T_m*m_T_b;
    Vector3d i_p_ib = i_T_b.translation();
    Matrix3d i_R_b = i_T_b.rotation();



    // initialize some vectors and fill them with dicretized data
    cnpy::npz_t arr_map = cnpy::npz_load(data_file_path);

    //load it into a new array
    cnpy::NpyArray t_npyarr = arr_map["t"];
    cnpy::NpyArray imu_acc_npyarr = arr_map["imu_acc"];
    // cnpy::NpyArray o_a_oi_npyarr = arr_map["o_a_oi"];
    cnpy::NpyArray o_q_i_npyarr = arr_map["o_q_i"];
    cnpy::NpyArray i_omg_oi_npyarr = arr_map["i_omg_oi"];
    cnpy::NpyArray qa_npyarr = arr_map["qa"];
    cnpy::NpyArray dqa_npyarr = arr_map["dqa"];
    cnpy::NpyArray tau_npyarr = arr_map["tau"];
    cnpy::NpyArray l_forces_npyarr = arr_map["l_forces"];
    // cnpy::NpyArray w_pose_wm_npyarr = arr_map["w_pose_wm"];
    cnpy::NpyArray m_v_wm_npyarr = arr_map["m_v_wm"];
    cnpy::NpyArray w_v_wm_npyarr = arr_map["w_v_wm"];
    // cnpy::NpyArray o_R_i_npyarr = arr_map["o_R_i"];
    cnpy::NpyArray w_p_wm_npyarr = arr_map["w_p_wm"];
    cnpy::NpyArray w_q_m_npyarr = arr_map["w_q_m"];
    cnpy::NpyArray contact_npyarr = arr_map["contacts"];
    // cnpy::NpyArray b_p_bl_mocap_npyarr = arr_map["b_p_bl_mocap"];

    // cnpy::NpyArray w_R_m_npyarr = arr_map["w_R_m"];
    double* t_arr = t_npyarr.data<double>();
    double* imu_acc_arr = imu_acc_npyarr.data<double>();
    // double* o_a_oi_arr = o_a_oi_npyarr.data<double>();
    double* i_omg_oi_arr = i_omg_oi_npyarr.data<double>();
    double* qa_arr = qa_npyarr.data<double>();
    double* dqa_arr = dqa_npyarr.data<double>();
    double* tau_arr = tau_npyarr.data<double>();
    double* l_forces_arr = l_forces_npyarr.data<double>();
    // double* w_pose_wm_arr = w_pose_wm_npyarr.data<double>();
    double* m_v_wm_arr = m_v_wm_npyarr.data<double>();
    double* w_v_wm_arr = w_v_wm_npyarr.data<double>();
    double* o_q_i_arr = o_q_i_npyarr.data<double>();
    // double* o_R_i_arr = o_R_i_npyarr.data<double>();
    double* w_p_wm_arr = w_p_wm_npyarr.data<double>();
    double* w_q_m_arr = w_q_m_npyarr.data<double>();
    
    double* contact_arr = contact_npyarr.data<double>();
    // double* b_p_bl_mocap_arr = b_p_bl_mocap_npyarr.data<double>();

    // double* w_R_m_arr = w_R_m_npyarr.data<double>();
    unsigned int N = t_npyarr.shape[0];
    if (max_t > 0){
        N = N < int(max_t/dt) ? N : int(max_t/dt);
    }

    // Load the urdf model
    Model model;
    pinocchio::urdf::buildModel(robot_urdf, JointModelFreeFlyer(), model);
    std::cout << "model name: " << model.name << std::endl;
    Data data(model);
    std::vector<std::string> ee_names = {"FL_ANKLE", "FR_ANKLE", "HL_ANKLE", "HR_ANKLE"};
    unsigned int nbc = ee_names.size();

    // Recover end effector frame ids
    std::vector<int> ee_ids;
    std::cout << "Frame ids" << std::endl;
    for (auto ee_name: ee_names){
        ee_ids.push_back(model.getFrameId(ee_name));
        std::cout << ee_name << std::endl;
    }

    /////////////////////////
    // WOLF enters the scene
    // SETUP PROBLEM/SENSORS/PROCESSORS
    std::string bodydynamics_root_dir = _WOLF_BODYDYNAMICS_ROOT_DIR;

    ProblemPtr problem = Problem::create("POV", 3);

    // add a tree manager for sliding window optimization 
    ParserYaml parser = ParserYaml("demos/tree_manager.yaml", bodydynamics_root_dir);
    ParamsServer server = ParamsServer(parser.getParams());
    auto tree_manager = TreeManagerSlidingWindow::create("tree_manager", server);
    problem->setTreeManager(tree_manager);

    //////////////////////
    // CERES WRAPPER
    SolverCeresPtr solver = std::make_shared<SolverCeres>(problem);
    solver->getSolverOptions().max_num_iterations = max_num_iterations;
    solver->getSolverOptions().function_tolerance = func_tol; // 1e-6
    solver->getSolverOptions().gradient_tolerance = 1e-4*solver->getSolverOptions().function_tolerance;

    // solver->getSolverOptions().minimizer_type = ceres::TRUST_REGION;
    // solver->getSolverOptions().trust_region_strategy_type = ceres::DOGLEG;
    // solver->getSolverOptions().trust_region_strategy_type = ceres::LEVENBERG_MARQUARDT;

    // solver->getSolverOptions().minimizer_type = ceres::LINE_SEARCH;
    // solver->getSolverOptions().line_search_direction_type = ceres::LBFGS;


    //===================================================== INITIALIZATION
    // SENSOR + PROCESSOR IMU
    SensorBasePtr sen_imu_base = problem->installSensor("SensorImu", "SenImu", (Vector7d() << 0,0,0, 0,0,0,1).finished(), bodydynamics_root_dir + "/demos/sensor_imu_solo12.yaml");
    SensorImuPtr sen_imu = std::static_pointer_cast<SensorImu>(sen_imu_base);
    ProcessorBasePtr proc_ftp_base = problem->installProcessor("ProcessorImu", "imu pre-integrator", "SenImu", bodydynamics_root_dir + "/demos/processor_imu_solo12.yaml");
    ProcessorImuPtr proc_imu = std::static_pointer_cast<ProcessorImu>(proc_ftp_base);

    // SENSOR + PROCESSOR POINT FEET NOMOVE
    ParamsSensorPointFeetNomovePtr intr_pfn = std::make_shared<ParamsSensorPointFeetNomove>();
    intr_pfn->std_foot_nomove_ = std_foot_nomove;
    intr_pfn->std_altitude_ = std_altitude;
    intr_pfn->foot_radius_ = foot_radius;
    VectorXd extr;  // default size for dynamic vector is 0-0 -> what we need
    SensorBasePtr sen_pfnm_base = problem->installSensor("SensorPointFeetNomove", "SenPfnm", extr, intr_pfn);
    SensorPointFeetNomovePtr sen_pfnm = std::static_pointer_cast<SensorPointFeetNomove>(sen_pfnm_base);
    ParamsProcessorPointFeetNomovePtr params_pfnm = std::make_shared<ParamsProcessorPointFeetNomove>();
    params_pfnm->voting_active = false;
    params_pfnm->time_tolerance = 0.0;
    params_pfnm->max_time_vote = 0.0;
    ProcessorBasePtr proc_pfnm_base = problem->installProcessor("ProcessorPointFeetNomove", "PPointFeetNomove", sen_pfnm, params_pfnm);


    // SETPRIOR RETRO-ENGINEERING
    // We are not using setPrior because of processors initial captures problems so we have to
    // - Manually create the first KF and its pose and velocity priors
    // - call setOrigin on processors MotionProvider since the flag prior_is_set is not true when doing installProcessor (later)

    //////////////////////
    // COMPUTE INITIAL STATE
    TimeStamp t0(t_arr[0]);
    Eigen::Map<Vector3d> w_p_wm(w_p_wm_arr);
    Eigen::Map<Vector4d> w_qvec_m_init(w_q_m_arr);
    w_qvec_m_init.normalize();  // not close enough to wolf eps sometimes
    Quaterniond w_q_m_init(w_qvec_m_init);
    Vector3d w_p_wi_init = w_p_wm + w_q_m_init*m_p_mi;
    Quaterniond w_q_i_init = w_q_m_init*m_q_i;
    VectorComposite x_origin("POV", {w_p_wi_init, w_q_i_init.coeffs(), Vector3d::Zero()});
    VectorComposite std_origin("POV", {std_prior_p*Vector3d::Ones(), std_prior_o*Vector3d::Ones(), std_prior_v*Vector3d::Ones()});
    // FrameBasePtr KF1 = problem->setPriorFactor(x_origin, std_origin, t0);
    FrameBasePtr KF1 = problem->setPriorInitialGuess(x_origin, t0);  // if mocap used
    
    proc_imu->setOrigin(KF1);

    //////////////////////////////////////////
    // INITIAL BIAS FACTORS
    // Add absolute factor on Imu biases to simulate a fix()
    Vector6d std_abs_imu; std_abs_imu << std_abs_bias_acc*Vector3d::Ones(), std_abs_bias_gyro*Vector3d::Ones();
    Matrix6d Q_bi = std_abs_imu.array().square().matrix().asDiagonal();    
    CaptureBasePtr capbi0 = CaptureBase::emplace<CaptureBase>(KF1, "bi0", t0);
    FeatureBasePtr featbi0 = FeatureBase::emplace<FeatureBase>(capbi0, "bi0", bias_imu_prior, Q_bi);
    FactorBasePtr facbi0 = FactorBase::emplace<FactorBlockAbsolute>(featbi0, featbi0, KF1->getCaptureOf(sen_imu)->getSensorIntrinsic(), nullptr, false);   
    KF1->getCaptureOf(sen_imu)->getStateBlock('I')->setState(bias_imu_prior);
    sen_imu->getStateBlock('I')->setState(bias_imu_prior);


    // Allocate on the heap to prevent stack overflow for large datasets
    double* o_p_ob_fbk_carr = new double[3*N];
    double* o_q_b_fbk_carr = new double[4*N];
    double* o_v_ob_fbk_carr = new double[3*N];
    double* o_p_oi_fbk_carr = new double[3*N];
    double* o_q_i_fbk_carr = new double[4*N];
    double* o_v_oi_fbk_carr = new double[3*N];
    double* imu_bias_fbk_carr = new double[6*N];
    double* i_pose_im_fbk_carr = new double[7*N];

    // covariances computed on the fly
    std::vector<Vector3d> Qp_fbk_v; Qp_fbk_v.push_back(Vector3d::Zero());
    std::vector<Vector3d> Qo_fbk_v; Qo_fbk_v.push_back(Vector3d::Zero());
    std::vector<Vector3d> Qv_fbk_v; Qv_fbk_v.push_back(Vector3d::Zero());
    std::vector<Vector6d> Qbi_fbk_v; Qbi_fbk_v.push_back(Vector6d::Zero());
    std::vector<Vector6d> Qm_fbk_v; Qm_fbk_v.push_back(Vector6d::Zero());


    std::vector<Vector3d> i_omg_oi_v; 
    //////////////////////////////////////////

    unsigned int nb_kf = 1;
    for (unsigned int i=0; i < N; i++){
        TimeStamp ts(t_arr[i]);

        ////////////////////////////////////
        // Retrieve current state
        VectorComposite curr_state = problem->getState();
        Vector4d o_qvec_i_curr = curr_state['O']; 
        Quaterniond o_q_i_curr(o_qvec_i_curr);
        Vector3d o_v_oi_curr = curr_state['V']; 


        //////////////
        // PROCESS IMU
        //////////////
        // Get measurements
        // retrieve traj data
        Eigen::Map<Vector3d> imu_acc(imu_acc_arr+3*i);
        Eigen::Map<Vector3d> i_omg_oi(i_omg_oi_arr+3*i);
        // store i_omg_oi for later computation of o_v_ob from o_v_oi
        i_omg_oi_v.push_back(i_omg_oi);

        Vector6d imu_bias = sen_imu->getStateBlockDynamic('I', t_arr[i])->getState();
        Vector3d i_omg_oi_unbiased = (i_omg_oi - imu_bias.tail<3>());

        Vector6d acc_gyr_meas; acc_gyr_meas << imu_acc, i_omg_oi;
        CaptureImuPtr CImu = std::make_shared<CaptureImu>(ts, sen_imu, acc_gyr_meas, sen_imu->getNoiseCov());
        CImu->process();
        //////////////
        //////////////

        //////////////////////////////////
        // Kinematics + forces
        // retrieve traj data
        Eigen::Map<Matrix<double,12, 1>> tau(tau_arr+12*i);
        Eigen::Map<Matrix<double,12, 1>> qa(qa_arr+12*i);
        Eigen::Map<Matrix<double,12, 1>> dqa(dqa_arr+12*i);
        Eigen::Map<Matrix<double,4, 1>> contact_gtr(contact_arr+4*i);  // int conversion does not work!
        // std::cout << contact_gtr.transpose() << std::endl;
        // Eigen::Map<Matrix<double,12, 1>> b_p_bl_mocap(b_p_bl_mocap_arr+12*i);  // int conversion does not work!

        // Kinematics model
        qa = qa + delta_qa + alpha_qa.cwiseProduct(tau);
        
        Matrix3d o_R_i = Quaterniond(o_q_i_curr).toRotationMatrix();  // IMU internal filter

        // Or else, retrieve from preprocessed dataset
        Eigen::Map<Matrix<double,12, 1>> l_forces(l_forces_arr+12*i);

        Vector3d o_f_tot = Vector3d::Zero();
        std::vector<Vector3d> l_f_l_vec;
        std::vector<Vector3d> o_f_l_vec;
        std::vector<Vector3d> i_p_il_vec;
        std::vector<Vector4d> i_qvec_l_vec;
        // needing relative kinematics measurements
        VectorXd q_static(19); q_static << 0,0,0, 0,0,0,1, qa;
        VectorXd dq_static(18); dq_static << 0,0,0, 0,0,0, dqa;
        forwardKinematics(model, data, q_static, dq_static);
        updateFramePlacements(model, data);

        for (unsigned int i_ee=0; i_ee < nbc; i_ee++){
            auto b_T_l = data.oMf[ee_ids[i_ee]];

            auto i_T_l = i_T_b * b_T_l;
            Vector3d i_p_il = i_T_l.translation();            
            Matrix3d i_R_l = i_T_l.rotation();
            Vector4d i_qvec_l = Quaterniond(i_R_l).coeffs();
            i_p_il = i_p_il +  i_R_l*l_p_lg;

            Vector3d l_f_l = l_forces.segment(3*i_ee, 3);
            Vector3d o_f_l = o_R_i*i_R_l * l_f_l;     // for contact test

            l_f_l_vec.push_back(l_f_l);
            o_f_l_vec.push_back(o_f_l);
            i_p_il_vec.push_back(i_p_il);
            i_qvec_l_vec.push_back(i_qvec_l);
        }

        // Detect feet in contact
        int nb_feet_in_contact = 0;
        std::unordered_map<int, Vector7d> kin_incontact;
        // std::cout << "" << std::endl;
        for (unsigned int i_ee=0; i_ee<nbc; i_ee++){
            // basic contact detection based on normal foot vel, things break if no foot is in contact
            // if ((o_f_l_vec[i_ee](2) > fz_thresh) && (nb_feet_in_contact < nb_feet)){
            // if ((contact_gtr[i_ee] > 0.5) && (nb_feet_in_contact < nb_feet)){
            if (contact_gtr[i_ee] > 0.5){
                nb_feet_in_contact++;
                Vector7d i_pose_il; i_pose_il << i_p_il_vec[i_ee], i_qvec_l_vec[i_ee];
                kin_incontact.insert({i_ee, i_pose_il});
            }
        }

        CapturePointFeetNomovePtr CPF = std::make_shared<CapturePointFeetNomove>(ts, sen_pfnm, kin_incontact);
        CPF->process();

        // solve every new KF
        if (problem->getTrajectory()->size() > nb_kf ){
            std::string report = solver->solve(SolverManager::ReportVerbosity::BRIEF);

            // ////////////////////////////////
            // // FOR PLOTS
            // if (solver->iterations() == 1){
            //     std::cout << "ONLY ONE!" << std::endl;
            //     printFactorCosts(kf_last);
            //     for (int k=0; k < 5; k++){
            //         std::cout << "P E R T U B!" << std::endl;
            //         // problem->perturb(0.000001);
            //         kf_last->perturb();
            //         // std::cout << "kf_last P " << kf_last->getP()->getState().transpose() << std::endl; 
            //         report = solver->solve(SolverManager::ReportVerbosity::BRIEF);
            //         printFactorCosts(kf_last);
            //     }

            // }
            // ///////////////////////


            ////////////////////////////////
            // FOR REAL
            if (solver->iterations() == 1){
                // kf_last->perturb(0.000001);
                report = solver->solve(SolverManager::ReportVerbosity::BRIEF);
            }
            ///////////////////////

            std::cout << ts << "  ";
            std::cout << report << std::endl;
            //////////////////////////////

            // printFactorCosts(kf_last);


            // compute covariances of KF and capture stateblocks
            Eigen::MatrixXd Qp_fbk =  Eigen::Matrix3d::Identity();
            Eigen::MatrixXd Qo_fbk =  Eigen::Matrix3d::Identity();  // global or local?
            Eigen::MatrixXd Qv_fbk =  Eigen::Matrix3d::Identity();
            Eigen::MatrixXd Qbi_fbk = Eigen::Matrix6d::Identity();
            Eigen::MatrixXd Qmp_fbk =  Eigen::Matrix3d::Identity();  // extr mocap
            Eigen::MatrixXd Qmo_fbk =  Eigen::Matrix3d::Identity();  // extr mocap
            
            if (compute_cov)
                solver->computeCovariances(SolverManager::CovarianceBlocksToBeComputed::ALL);  // should not be computed every time !! see below
            problem->getCovarianceBlock(kf_last->getStateBlock('P'), Qp_fbk);
            problem->getCovarianceBlock(kf_last->getStateBlock('O'), Qo_fbk);
            problem->getCovarianceBlock(kf_last->getStateBlock('V'), Qv_fbk);

            CaptureBasePtr cap_imu = kf_last->getCaptureOf(sen_imu); 
            if (compute_cov)
                solver->computeCovariances(cap_imu->getStateBlockVec());  // not computed by ALL and destroys other computed covs -> issue to raise
            problem->getCovarianceBlock(cap_imu->getSensorIntrinsic(), Qbi_fbk);

            // Retrieve diagonals
            Vector3d Qp_fbk_diag = Qp_fbk.diagonal(); 
            Vector3d Qo_fbk_diag = Qo_fbk.diagonal(); 
            Vector3d Qv_fbk_diag = Qv_fbk.diagonal(); 
            Vector6d Qbi_fbk_diag = Qbi_fbk.diagonal(); 
            Vector6d Qm_fbk_diag; Qm_fbk_diag << Qmp_fbk.diagonal(), Qmo_fbk.diagonal(); 
            Vector3d Qmo_fbk_diag = Qmo_fbk.diagonal(); 

            Qp_fbk_v.push_back(Qp_fbk_diag);
            Qo_fbk_v.push_back(Qo_fbk_diag);
            Qv_fbk_v.push_back(Qv_fbk_diag);
            Qbi_fbk_v.push_back(Qbi_fbk_diag);
            Qm_fbk_v.push_back(Qm_fbk_diag);

            nb_kf = problem->getTrajectory()->getFrameMap().size();
        }

        // Store current state estimation
        VectorComposite state_fbk = problem->getState(ts);
        Vector3d o_p_oi = state_fbk['P'];
        Vector4d o_qvec_i = state_fbk['O'];
        Vector3d o_v_oi = state_fbk['V'];

        // IMU to base frame 
        o_R_i = Quaterniond(o_qvec_i).toRotationMatrix();
        Matrix3d o_R_b = o_R_i * i_R_b;
        Vector4d o_qvec_b = Quaterniond(o_R_b).coeffs();
        Vector3d o_p_ob = o_p_oi + o_R_i * i_p_ib;
        // Vector3d o_v_ob = o_v_oi + o_R_i * i_omg_oi_unbiased.cross(i_p_ib);
        Vector3d o_v_ob = o_v_oi + o_R_i * i_omg_oi.cross(i_p_ib);

        imu_bias = sen_imu->getIntrinsic()->getState();
        Vector7d i_pose_im_fbk; i_pose_im_fbk << i_p_im, i_qvec_m;

        mempcpy(o_p_ob_fbk_carr+3*i, o_p_ob.data(),             3*sizeof(double));
        mempcpy(o_q_b_fbk_carr +4*i, o_qvec_b.data(),           4*sizeof(double));
        mempcpy(o_v_ob_fbk_carr+3*i, o_v_ob.data(),             3*sizeof(double));
        mempcpy(o_p_oi_fbk_carr+3*i, o_p_oi.data(),             3*sizeof(double));
        mempcpy(o_q_i_fbk_carr +4*i, o_qvec_i.data(),           4*sizeof(double));
        mempcpy(o_v_oi_fbk_carr+3*i, o_v_oi.data(),             3*sizeof(double));
        mempcpy(imu_bias_fbk_carr+6*i, imu_bias.data(),         6*sizeof(double));
        mempcpy(i_pose_im_fbk_carr+7*i, i_pose_im_fbk.data(),   7*sizeof(double));
    }
    


    ///////////////////////////////////////////
    //////////////// SOLVING //////////////////
    ///////////////////////////////////////////
    // if (solve_end){
    //     std::string report = solver->solve(SolverManager::ReportVerbosity::BRIEF);
    //     problem->print(3,true,false,false);
    //     std::cout << report << std::endl;
    // }

    double* o_p_ob_carr = new double[3*N];
    double* o_q_b_carr = new double[4*N];
    double* o_v_ob_carr = new double[3*N];
    double* o_p_oi_carr = new double[3*N];
    double* o_q_i_carr = new double[4*N];
    double* o_v_oi_carr = new double[3*N];
    double* imu_bias_carr = new double[6*N];

    for (unsigned int i=0; i < N; i++) { 
        VectorComposite state_est = problem->getState(t_arr[i]);
        Vector3d i_omg_oi = i_omg_oi_v[i];
        Vector3d o_p_oi = state_est['P'];
        Vector4d o_qvec_i = state_est['O'];
        Vector3d o_v_oi = state_est['V'];

        auto sb = sen_imu->getStateBlockDynamic('I', t_arr[i]);
        Vector6d imu_bias = sb->getState();
        Vector3d i_omg_oi_unbiased = (i_omg_oi - imu_bias.tail<3>());

        Matrix3d o_R_i = Quaterniond(o_qvec_i).toRotationMatrix();
        Matrix3d o_R_b = o_R_i * i_R_b;
        Vector4d o_qvec_b = Quaterniond(o_R_b).coeffs();
        Vector3d o_p_ob = o_p_oi + o_R_i * i_p_ib;
        // Vector3d o_v_ob = o_v_oi + o_R_i * i_omg_oi_unbiased.cross(i_p_ib);
        Vector3d o_v_ob = o_v_oi + o_R_i * i_omg_oi.cross(i_p_ib);

        mempcpy(o_p_ob_carr+3*i, o_p_ob.data(),     3*sizeof(double));
        mempcpy(o_q_b_carr +4*i, o_qvec_b.data(),   4*sizeof(double));
        mempcpy(o_v_ob_carr+3*i, o_v_ob.data(),     3*sizeof(double));
        mempcpy(o_p_oi_carr+3*i, o_p_oi.data(),     3*sizeof(double));
        mempcpy(o_q_i_carr +4*i, o_qvec_i.data(),   4*sizeof(double));
        mempcpy(o_v_oi_carr+3*i, o_v_oi.data(),     3*sizeof(double));
        mempcpy(imu_bias_carr+6*i, imu_bias.data(), 6*sizeof(double));
    }


    // Compute Covariances
    unsigned int Nkf = problem->getTrajectory()->size();
    double* tkf_carr  = new double[1*Nkf];
    double* Qp_carr  = new double[3*Nkf];
    double* Qo_carr  = new double[3*Nkf];
    double* Qv_carr  = new double[3*Nkf];
    double* Qbi_carr = new double[6*Nkf];
    double* Qm_carr  = new double[6*Nkf];
    // feedback covariances
    double* Qp_fbk_carr  = new double[3*Nkf];
    double* Qo_fbk_carr  = new double[3*Nkf];
    double* Qv_fbk_carr  = new double[3*Nkf];
    double* Qbi_fbk_carr = new double[6*Nkf];
    double* Qm_fbk_carr  = new double[6*Nkf];

    // factor errors
    double* fac_imu_err_carr = new double[9*Nkf];
    double* fac_pose_err_carr  = new double[6*Nkf];
    int i = 0;
    for (auto elt: problem->getTrajectory()->getFrameMap())
    {
        std::cout << "Traj " << i << "/" << problem->getTrajectory()->size() << std::endl;
        tkf_carr[i] = elt.first.get();
        auto kf = elt.second;
        VectorComposite kf_state = kf->getState();
        
        // compute covariances of KF and capture stateblocks
        Eigen::MatrixXd Qp =  Eigen::Matrix3d::Identity();
        Eigen::MatrixXd Qo =  Eigen::Matrix3d::Identity();  // global or local?
        Eigen::MatrixXd Qv =  Eigen::Matrix3d::Identity();
        Eigen::MatrixXd Qbi = Eigen::Matrix6d::Identity();
        Eigen::MatrixXd Qmp =  Eigen::Matrix3d::Identity();  // extr mocap
        Eigen::MatrixXd Qmo =  Eigen::Matrix3d::Identity();  // extr mocap
        
        if (compute_cov)
            solver->computeCovariances(SolverManager::CovarianceBlocksToBeComputed::ALL);  // should not be computed every time !! see below
        problem->getCovarianceBlock(kf->getStateBlock('P'), Qp);
        problem->getCovarianceBlock(kf->getStateBlock('O'), Qo);
        problem->getCovarianceBlock(kf->getStateBlock('V'), Qv);

        CaptureBasePtr cap_imu = kf->getCaptureOf(sen_imu); 
        if (compute_cov)
            solver->computeCovariances(cap_imu->getStateBlockVec());  // not computed by ALL and destroys other computed covs -> issue to raise
        problem->getCovarianceBlock(cap_imu->getSensorIntrinsic(), Qbi);

        // diagonal components
        Vector3d Qp_diag = Qp.diagonal(); 
        Vector3d Qo_diag = Qo.diagonal(); 
        Vector3d Qv_diag = Qv.diagonal(); 
        Vector6d Qbi_diag = Qbi.diagonal(); 
        Vector6d Qm_diag; Qm_diag << Qmp.diagonal(), Qmo.diagonal(); 
        
        memcpy(Qp_carr  + 3*i, Qp_diag.data(),  3*sizeof(double)); 
        memcpy(Qo_carr  + 3*i, Qo_diag.data(),  3*sizeof(double)); 
        memcpy(Qv_carr  + 3*i, Qv_diag.data(),  3*sizeof(double)); 
        memcpy(Qbi_carr + 6*i, Qbi_diag.data(), 6*sizeof(double)); 
        memcpy(Qm_carr  + 6*i, Qm_diag.data(),  6*sizeof(double)); 

        // Recover feedback covariances
        memcpy(Qp_fbk_carr  + 3*i, Qp_fbk_v[i].data(), 3*sizeof(double)); 
        memcpy(Qo_fbk_carr  + 3*i, Qo_fbk_v[i].data(), 3*sizeof(double)); 
        memcpy(Qv_fbk_carr  + 3*i, Qv_fbk_v[i].data(), 3*sizeof(double)); 
        memcpy(Qbi_fbk_carr + 6*i, Qbi_fbk_v[i].data(), 6*sizeof(double)); 
        memcpy(Qm_fbk_carr + 6*i, Qm_fbk_v[i].data(), 6*sizeof(double)); 
        

        // Factor errors
        // std::cout << "Factors " << kf->getTimeStamp().get() << std::endl;
        FactorBasePtrList fac_lst; 
        kf->getFactorList(fac_lst);
        Vector9d imu_err = Vector9d::Zero();
        Vector6d pose_err = Vector6d::Zero();

        i++;
    }


    // Save trajectories in npz file
    // save ground truth    
    cnpy::npz_save(out_npz_file_path, "t", t_arr, {N}, "w");             // "w" overwrites any file with same name
    cnpy::npz_save(out_npz_file_path, "w_p_wm", w_p_wm_arr, {N,3}, "a"); // "a" appends to the file we created above
    cnpy::npz_save(out_npz_file_path, "w_q_m",  w_q_m_arr,  {N,4}, "a");
    cnpy::npz_save(out_npz_file_path, "w_v_wm", w_v_wm_arr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "m_v_wm", m_v_wm_arr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_q_i_imu",  o_q_i_arr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "qa",     qa_arr, {N,12}, "a");
    cnpy::npz_save(out_npz_file_path, "i_omg_oi", i_omg_oi_arr, {N,3}, "a");

    // Online estimates
    cnpy::npz_save(out_npz_file_path, "o_p_ob_fbk", o_p_ob_fbk_carr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_q_b_fbk",  o_q_b_fbk_carr,  {N,4}, "a");
    cnpy::npz_save(out_npz_file_path, "o_v_ob_fbk", o_v_ob_fbk_carr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_p_oi_fbk", o_p_oi_fbk_carr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_q_i_fbk",  o_q_i_fbk_carr,  {N,4}, "a");
    cnpy::npz_save(out_npz_file_path, "o_v_oi_fbk", o_v_oi_fbk_carr, {N,3}, "a");
    
    // offline states
    cnpy::npz_save(out_npz_file_path, "o_p_ob", o_p_ob_carr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_q_b",  o_q_b_carr,  {N,4}, "a");
    cnpy::npz_save(out_npz_file_path, "o_v_ob", o_v_ob_carr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_p_oi", o_p_oi_carr, {N,3}, "a");
    cnpy::npz_save(out_npz_file_path, "o_q_i",  o_q_i_carr,  {N,4}, "a");
    cnpy::npz_save(out_npz_file_path, "o_v_oi", o_v_oi_carr, {N,3}, "a");

    // and biases/extrinsics
    cnpy::npz_save(out_npz_file_path, "imu_bias_fbk", imu_bias_fbk_carr, {N,6}, "a");
    cnpy::npz_save(out_npz_file_path, "imu_bias", imu_bias_carr, {N,6}, "a");
    cnpy::npz_save(out_npz_file_path, "i_pose_im_fbk", i_pose_im_fbk_carr, {N,7}, "a");

    // covariances
    cnpy::npz_save(out_npz_file_path, "tkf", tkf_carr, {Nkf}, "a");
    cnpy::npz_save(out_npz_file_path, "Qp", Qp_carr,   {Nkf,3}, "a");
    cnpy::npz_save(out_npz_file_path, "Qo", Qo_carr,   {Nkf,3}, "a");
    cnpy::npz_save(out_npz_file_path, "Qv", Qv_carr,   {Nkf,3}, "a");
    cnpy::npz_save(out_npz_file_path, "Qbi", Qbi_carr, {Nkf,6}, "a");
    cnpy::npz_save(out_npz_file_path, "Qm", Qm_carr,   {Nkf,6}, "a");
    cnpy::npz_save(out_npz_file_path, "Qp_fbk", Qp_fbk_carr, {Nkf,3}, "a");
    cnpy::npz_save(out_npz_file_path, "Qo_fbk", Qo_fbk_carr, {Nkf,3}, "a");
    cnpy::npz_save(out_npz_file_path, "Qv_fbk", Qv_fbk_carr, {Nkf,3}, "a");
    cnpy::npz_save(out_npz_file_path, "Qbi_fbk", Qbi_fbk_carr, {Nkf,6}, "a");
    cnpy::npz_save(out_npz_file_path, "Qm_fbk", Qm_fbk_carr, {Nkf,6}, "a");
    cnpy::npz_save(out_npz_file_path, "fac_imu_err", fac_imu_err_carr, {Nkf,9}, "a");
    cnpy::npz_save(out_npz_file_path, "fac_pose_err", fac_pose_err_carr, {Nkf,6}, "a");

}
