//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#ifndef FACTOR_FORCE_TORQUE_INERTIAL_DYNAMICS_H
#define FACTOR_FORCE_TORQUE_INERTIAL_DYNAMICS_H

#include "bodydynamics/math/force_torque_inertial_delta_tools.h"
#include <core/feature/feature_motion.h>
#include <core/factor/factor_autodiff.h>

namespace wolf
{
using namespace Eigen;
using namespace bodydynamics;

WOLF_PTR_TYPEDEFS(FactorForceTorqueInertialDynamics);

/**
 * @brief Pre-integrated factor taking IMU and force-torque measurements (FTI)
 *
 * This factor evaluates the preintegrated FTI against pose, linear velocity, IMU bias and angular momentum.
 *
 * State blocks considered:
 *  - Frame 1 (origin)
 *    'P' position
 *    'V' velocity
 *    'O' orientation
 *    'L' angular momentum
 *  - Capture 1 (origin)
 *    'I' IMU biases
 *  - Frame 2
 *    'P' position
 *    'V' velocity
 *    'O' orientation
 *    'L' angular momentum
 *  - Sensor Force Torque Inertial
 *    'C' position of the center of mass
 *    'i' inertia matrix components (we are assuming that it is a diagonal matrix)
 *    'm' mass
 *
 * The residual computed has the following components, in this order
 *  - position delta error according to IMU preintegration
 *  - velocity delta error according to IMU preintegration
 *  - position delta error according to FT preintegration
 *  - velocity delta error according to FT preintegration
 *  - angular momentum delta error according to FT preintegration
 *  - orientation delta error according to IMU preintegration
 */
class FactorForceTorqueInertialDynamics
    : public FactorAutodiff<FactorForceTorqueInertialDynamics, 18, 3, 4, 3, 3, 6, 3, 4, 3, 3, 3, 3, 1>  // POVLB - POVL
                                                                                                        // - Cim
                                                                                                        // // TODO: add
                                                                                                        // FT bias?
{
  public:
    FactorForceTorqueInertialDynamics(const FeatureMotionPtr& _ftr,  // gets measurement and access to parents
                                      const CaptureBasePtr&   _capture_origin,  // gets biases
                                      const ProcessorBasePtr& _processor,       // gets access to processor and sensor
                                      bool                    _apply_loss_function,
                                      FactorStatus            _status = FAC_ACTIVE);

    ~FactorForceTorqueInertialDynamics() override = default;

    template <typename T>
    bool operator()(const T* const _p1,  // position 1
                    const T* const _q1,  // orientation
                    const T* const _v1,  // velocity
                    const T* const _L1,  // ang momentum
                    const T* const _b1,  // IMU bias (acc and gyro)
                    const T* const _p2,  // position 2
                    const T* const _q2,  // orientation
                    const T* const _v2,  // velocity
                    const T* const _L2,  // ang momentum
                    const T* const _C,   // center of mass
                    const T* const _i,   // inertia matrix
                    const T* const _m,   // mass
                    T*             _res) const;      // residual

    template <typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, typename D7, typename D8>
    bool residual(const MatrixBase<D1>&     _p1,
                  const QuaternionBase<D4>& _q1,
                  const MatrixBase<D1>&     _v1,
                  const MatrixBase<D1>&     _L1,
                  const MatrixBase<D2>&     _b1,
                  const MatrixBase<D3>&     _p2,
                  const QuaternionBase<D5>& _q2,
                  const MatrixBase<D3>&     _v2,
                  const MatrixBase<D3>&     _L2,
                  const MatrixBase<D6>&     _C,
                  const MatrixBase<D6>&     _i,
                  const D7&                 _m,
                  MatrixBase<D8>&           _res) const;

    VectorXd residual() const;

    VectorXd getCalibPreint() const
    {
        return calib_preint_;
    }

  private:
    double                 dt_;
    Matrix<double, 19, 1>  delta_preint_;
    Matrix<double, 13, 1>  calib_preint_;
    Matrix<double, 18, 13> J_delta_calib_;
    Matrix<double, 18, 18> sqrt_info_upper_;
};

inline FactorForceTorqueInertialDynamics::FactorForceTorqueInertialDynamics(const FeatureMotionPtr& _ftr,
                                                                            const CaptureBasePtr&   _capture_origin,
                                                                            const ProcessorBasePtr& _processor,
                                                                            bool         _apply_loss_function,
                                                                            FactorStatus _status)
    : FactorAutodiff<FactorForceTorqueInertialDynamics, 18, 3, 4, 3, 3, 6, 3, 4, 3, 3, 3, 3, 1>(
          "FactorForceTorqueInertialDynamics",
          TOP_MOTION,
          _ftr,                         // parent feature
          _capture_origin->getFrame(),  // frame other
          _capture_origin,              // capture other
          nullptr,                      // feature other
          nullptr,                      // landmark other
          _processor,                   // processor
          _apply_loss_function,
          _status,
          _capture_origin->getFrame()->getStateBlock('P'),    // previous frame P
          _capture_origin->getFrame()->getStateBlock('O'),    // previous frame O
          _capture_origin->getFrame()->getStateBlock('V'),    // previous frame V
          _capture_origin->getFrame()->getStateBlock('L'),    // previous frame L
          _capture_origin->getStateBlock('I'),                // previous frame IMU bias
          _ftr->getFrame()->getStateBlock('P'),               // current frame P
          _ftr->getFrame()->getStateBlock('O'),               // current frame O
          _ftr->getFrame()->getStateBlock('V'),               // current frame V
          _ftr->getFrame()->getStateBlock('L'),               // current frame L
          _capture_origin->getSensor()->getStateBlock('C'),   // sensor C
          _capture_origin->getSensor()->getStateBlock('i'),   // sensor i
          _capture_origin->getSensor()->getStateBlock('m')),  // sensor m
      dt_(_ftr->getFrame()->getTimeStamp() - _capture_origin->getFrame()->getTimeStamp()),
      delta_preint_(_ftr->getDeltaPreint()),
      calib_preint_(_ftr->getCalibrationPreint()),
      J_delta_calib_(_ftr->getJacobianCalibration()),
      sqrt_info_upper_(_ftr->getMeasurementSquareRootInformationUpper())
{
    //
}

template <typename D1, typename D2, typename D3, typename D4, typename D5, typename D6, typename D7, typename D8>
inline bool FactorForceTorqueInertialDynamics::residual(const MatrixBase<D1>&     _p1,
                                                        const QuaternionBase<D4>& _q1,
                                                        const MatrixBase<D1>&     _v1,
                                                        const MatrixBase<D1>&     _L1,
                                                        const MatrixBase<D2>&     _b1,
                                                        const MatrixBase<D3>&     _p2,
                                                        const QuaternionBase<D5>& _q2,
                                                        const MatrixBase<D3>&     _v2,
                                                        const MatrixBase<D3>&     _L2,
                                                        const MatrixBase<D6>&     _C,
                                                        const MatrixBase<D6>&     _i,
                                                        const D7&                  _m,
                                                        MatrixBase<D8>&           _res) const
{
    MatrixSizeCheck<18, 1>::check(_res);

    typedef typename D4::Scalar  T;
    typedef Map<Matrix<T, 3, 1>> Matrixt;

    /* Will do the following:
     *
     * Compute estimated delta, using betweenStates()
     *    d_est = x2 (-) x1
     *
     * Compute correction step according to bias change, this is a linear operation
     *    d_step = J_delta_calib * (calib - bias_preint)
     *
     * Correct preintegrated delta with new bias, using plus()
     *    d_corr = d_preint (+) d_step
     *
     * Compute error, using diff()
     *    d_err = d_corr (-) d_est
     *
     * Compute residual
     *    res = W * d_err , with W the sqrt of the info matrix.
     */

    /* WARNING -- order of state blocks might be confusing!!
     * Frame blocks are ordered as follows:  "POVL"         -- pos, ori, vel, ang_momentum
     * Factor blocks are ordered as follows: "POVL-B-POVL"  -- B is IMU bias
     * Delta blocks are ordered as follows:  "PVpvLO"       -- where "P,V,O" are from IMU, and "p,v,L" from FT
     */

    Matrix<double, 3, 1> f;
    Matrix<T, 19, 1>     delta_est;           // delta_est = x2 (-) x1, computed with betweenStates(x1,x2,dt)
    Matrixt              dpi(&delta_est(0));  // 'P'
    Matrixt              dvi(&delta_est(3));  // 'V'
    Matrixt              dpd(&delta_est(6));  // 'p'
    Matrixt              dvd(&delta_est(9));  // 'v'
    Matrixt              dL(&delta_est(12));  // 'L'
    Map<Quaternion<T>>   dq(&delta_est(15));  // 'O'
    fti::betweenStates(_p1, _v1, _L1, _q1, _p2, _v2, _L2, _q2, dt_, dpi, dvi, dpd, dvd, dL, dq);

    Matrix<T, 19, 1> delta_preint = delta_preint_.cast<T>();
    Matrix<T, 13, 1> calib;
    calib << _b1, _C, _i, _m;
    Matrix<T, 18, 1> delta_step = J_delta_calib_ * (calib - calib_preint_);
    Matrix<T, 19, 1> delta_corr = fti::plus(delta_preint, delta_step);
    Matrix<T, 18, 1> delta_err  = fti::diff(delta_est, delta_corr);
    _res                        = sqrt_info_upper_ * delta_err;

    return true;
}

inline VectorXd FactorForceTorqueInertialDynamics::residual() const
{
    VectorXd res(18);
    Vector3d p0 = getFrameOther()->getStateBlock('P')->getState(); // previous frame P
    Quaterniond q0 = Quaterniond(getFrameOther()->getStateBlock('O')->getState().data());
    Vector3d v0 = getFrameOther()->getStateBlock('V')->getState();
    Vector3d L0 = getFrameOther()->getStateBlock('L')->getState();
    Vector6d bias = getCaptureOther()->getSensorIntrinsic()->getState();
    Vector3d p1 = getFrame()->getStateBlock('P')->getState(); // previous frame P
    Quaterniond q1 = Quaterniond(getFrame()->getStateBlock('O')->getState().data());
    Vector3d v1 = getFrame()->getStateBlock('V')->getState();
    Vector3d L1 = getFrame()->getStateBlock('L')->getState();
    Vector3d C = getCapture()->getSensor()->getStateBlock('C')->getState();
    Vector3d i = getCapture()->getSensor()->getStateBlock('i')->getState();
    double m = getCapture()->getSensor()->getStateBlock('m')->getState()(0);

    residual(p0, q0, v0, L0, bias, p1, q1, v1, L1, C, i, m, res); 
    return res;
}

template <typename T>
inline bool FactorForceTorqueInertialDynamics::operator()(const T* const _p1,
                                                          const T* const _q1,
                                                          const T* const _v1,
                                                          const T* const _L1,
                                                          const T* const _b1,
                                                          const T* const _p2,
                                                          const T* const _q2,
                                                          const T* const _v2,
                                                          const T* const _L2,
                                                          const T* const _C,
                                                          const T* const _i,
                                                          const T* const _m,
                                                          T* _res) const
{
    Map<const Matrix<T, 3, 1>> p1(_p1);
    Map<const Quaternion<T>>   q1(_q1);
    Map<const Matrix<T, 3, 1>> v1(_v1);
    Map<const Matrix<T, 3, 1>> L1(_L1);
    Map<const Matrix<T, 6, 1>> b1(_b1);
    Map<const Matrix<T, 3, 1>> p2(_p2);
    Map<const Quaternion<T>>   q2(_q2);
    Map<const Matrix<T, 3, 1>> v2(_v2);
    Map<const Matrix<T, 3, 1>> L2(_L2);
    Map<const Matrix<T, 3, 1>> C(_C);
    Map<const Matrix<T, 3, 1>> i(_i);
    const T&                   m = *_m;
    Map<Matrix<T, 18, 1>>      res(_res);

    residual(p1, q1, v1, L1, b1, p2, q2, v2, L2, C, i, m, res);

    return true;
}

}  // namespace wolf

#endif  // FACTOR_FORCE_TORQUE_INERTIAL_H