//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#ifndef FACTOR_FORCE_TORQUE_PREINT_THETA_H_
#define FACTOR_FORCE_TORQUE_PREINT_THETA_H_

//Wolf includes
#include "bodydynamics/capture/capture_force_torque_preint.h"
#include "bodydynamics/feature/feature_force_torque_preint.h"
#include "bodydynamics/sensor/sensor_force_torque.h"
#include "core/factor/factor_autodiff.h"
#include "core/math/rotations.h"

//Eigen include

namespace wolf {
    
WOLF_PTR_TYPEDEFS(FactorForceTorquePreint);

//class
class FactorForceTorquePreint : public FactorAutodiff<FactorForceTorquePreint, 12, 3,3,3,4,3,6, 3,3,3,4>
{
    public:
        FactorForceTorquePreint(const FeatureForceTorquePreintPtr& _ftr_ptr,
                                const CaptureForceTorquePreintPtr& _cap_origin_ptr, // gets to bp1, bw1
                                const ProcessorBasePtr&            _processor_ptr,
                                const CaptureBasePtr&               _cap_ikin_other,
                                const CaptureBasePtr&               _cap_gyro_other,
                                bool                               _apply_loss_function,
                                FactorStatus                       _status = FAC_ACTIVE);

        ~FactorForceTorquePreint() override = default;

        /** \brief : compute the residual from the state blocks being iterated by the solver.
            -> computes the expected measurement
            -> corrects actual measurement with new bias
            -> compares the corrected measurement with the expected one
            -> weights the result with the covariance of the noise (residual = sqrt_info_matrix * err;)
        */
        template<typename T>
        bool operator ()(const T* const _c1,
                         const T* const _cd1,
                         const T* const _Lc1,
                         const T* const _q1,
                         const T* const _bp1,
                         const T* const _bw1,
                         const T* const _c2,
                         const T* const _cd2,
                         const T* const _Lc2,
                         const T* const _q2,
                         T* _res) const;

        /** \brief : compute the residual from the state blocks being iterated by the solver. (same as operator())
            -> computes the expected measurement
            -> corrects actual measurement with new bias
            -> compares the corrected measurement with the expected one
            -> weights the result with the covariance of the noise (residual = sqrt_info_matrix * err;)
         * params :
         * _c1 : COM position at time t1 in world frame
         * _cd1: COM velocity at time t1 in world frame
         * _Lc1: Angular momentum at time t1 in world frame
         * _q1 : Base orientation at time t1
         * _bp1: COM position measurement  at time t1 in body frame
         * _bw1: gyroscope bias at time t1 in body frame 
         * _c2 : COM position at time t2 in world frame
         * _cd2: COM velocity at time t2 in world frame
         * _Lc2: Angular momentum at time t2 in world frame
         * _q2 : Base orientation at time t2
         * _res: Factor residuals (c,cd,Lc,o) O is rotation vector... NOT A QUATERNION
        */
        template<typename D1, typename D2, typename D3, typename D4>
        bool residual(const MatrixBase<D1> &     _c1,
                      const MatrixBase<D1> &     _cd1,
                      const MatrixBase<D1> &     _Lc1,
                      const QuaternionBase<D2> & _q1,
                      const MatrixBase<D1> &     _bp1,
                      const MatrixBase<D1> &     _bw1,
                      const MatrixBase<D1> &     _c2,
                      const MatrixBase<D1> &     _cd2,
                      const MatrixBase<D1> &     _Lc2,
                      const QuaternionBase<D3> & _q2,
                      MatrixBase<D4> &           _res) const;

        // Matrix<double,12,1> residual() const;
        // double cost() const;

    private:
        /// Preintegrated delta
        Vector3d    dc_preint_;
        Vector3d    dcd_preint_;
        Vector3d    dLc_preint_;
        Quaterniond dq_preint_;

        // Biases used during preintegration
        Vector3d pbc_bias_preint_;
        Vector3d gyro_bias_preint_;

        // Jacobians of preintegrated deltas wrt biases
        Matrix3d J_dLc_pb_;
        Matrix3d J_dc_wb_;
        Matrix3d J_dcd_wb_;
        Matrix3d J_dLc_wb_;
        Matrix3d J_dq_wb_;

        /// Metrics
        double dt_; ///< delta-time and delta-time-squared between keyframes
        double mass_; ///< constant total robot mass
        
};

///////////////////// IMPLEMENTATION ////////////////////////////

inline FactorForceTorquePreint::FactorForceTorquePreint(
                            const FeatureForceTorquePreintPtr& _ftr_ptr,
                            const CaptureForceTorquePreintPtr& _cap_origin_ptr,
                            const ProcessorBasePtr&            _processor_ptr,
                            const CaptureBasePtr&              _cap_ikin_other,
                            const CaptureBasePtr&              _cap_gyro_other,
                            bool                               _apply_loss_function,
                            FactorStatus                       _status) :
                FactorAutodiff<FactorForceTorquePreint, 12, 3,3,3,4,3,6, 3,3,3,4>(
                        "FactorForceTorquePreint",
                        TOP_MOTION,
                        _ftr_ptr,
                        _cap_origin_ptr->getFrame(),
                        _cap_origin_ptr,
                        nullptr,
                        nullptr,
                        _processor_ptr,
                        _apply_loss_function,
                        _status,
                        _cap_origin_ptr->getFrame()->getStateBlock('C'),
                        _cap_origin_ptr->getFrame()->getStateBlock('D'),
                        _cap_origin_ptr->getFrame()->getStateBlock('L'),
                        _cap_origin_ptr->getFrame()->getO(),
                        _cap_ikin_other->getSensorIntrinsic(),
                        _cap_gyro_other->getSensorIntrinsic(),
                        _ftr_ptr->getFrame()->getStateBlock('C'),
                        _ftr_ptr->getFrame()->getStateBlock('D'),
                        _ftr_ptr->getFrame()->getStateBlock('L'),
                        _ftr_ptr->getFrame()->getO()
                        ),
        dc_preint_(_ftr_ptr->getMeasurement().head<3>()), // dc, dcd, dLc, dq at preintegration time
        dcd_preint_(_ftr_ptr->getMeasurement().segment<3>(3)),
        dLc_preint_(_ftr_ptr->getMeasurement().segment<3>(6)),
        dq_preint_(_ftr_ptr->getMeasurement().data()+9),
        pbc_bias_preint_( std::static_pointer_cast<FeatureForceTorquePreint>(_ftr_ptr)->getPbcBiasPreint()),
        gyro_bias_preint_(std::static_pointer_cast<FeatureForceTorquePreint>(_ftr_ptr)->getGyroBiasPreint()),
        J_dLc_pb_(_ftr_ptr->getJacobianBias().block<3,3>(6,0)), // Jac dLc wrt pbc bias
        J_dc_wb_ (_ftr_ptr->getJacobianBias().block<3,3>(0,3)), // Jac dc wrt gyro bias
        J_dcd_wb_(_ftr_ptr->getJacobianBias().block<3,3>(3,3)), // Jac dc wrt gyro bias
        J_dLc_wb_(_ftr_ptr->getJacobianBias().block<3,3>(6,3)), // Jac dcd wrt gyro bias
        J_dq_wb_ (_ftr_ptr->getJacobianBias().block<3,3>(9,3)), // Jac dLc wrt gyro bias
        dt_(_ftr_ptr->getFrame()->getTimeStamp() - _cap_origin_ptr->getFrame()->getTimeStamp()),
        mass_(std::static_pointer_cast<SensorForceTorque>(_ftr_ptr->getCapture()->getSensor())->getMass())
{
//
}

template<typename T>
inline bool FactorForceTorquePreint::operator ()(const T* const _c1,
                                                 const T* const _cd1,
                                                 const T* const _Lc1,
                                                 const T* const _q1,
                                                 const T* const _bp1,
                                                 const T* const _bw1,
                                                 const T* const _c2,
                                                 const T* const _cd2,
                                                 const T* const _Lc2,
                                                 const T* const _q2,
                                                 T* _res) const
{
    Map<const Matrix<T,3,1> > c1(_c1);
    Map<const Matrix<T,3,1> > cd1(_cd1);
    Map<const Matrix<T,3,1> > Lc1(_Lc1);
    Map<const Quaternion<T> > q1(_q1);
    Map<const Matrix<T,3,1> > bp1(_bp1);
    Map<const Matrix<T,3,1> > bw1(_bw1 + 3);  // bw1 = bimu = [ba, bw]
    Map<const Matrix<T,3,1> > c2(_c2);
    Map<const Matrix<T,3,1> > cd2(_cd2);
    Map<const Matrix<T,3,1> > Lc2(_Lc2);
    Map<const Quaternion<T> > q2(_q2);
    Map<Matrix<T,12,1> > res(_res);

    residual(c1, cd1, Lc1, q1, bp1, bw1, c2, cd2, Lc2, q2, res);

    return true;
}

template<typename D1, typename D2, typename D3, typename D4>
inline bool FactorForceTorquePreint::residual(const MatrixBase<D1> &     _c1,
                                              const MatrixBase<D1> &     _cd1,
                                              const MatrixBase<D1> &     _Lc1,
                                              const QuaternionBase<D2> & _q1,
                                              const MatrixBase<D1> &     _bp1,
                                              const MatrixBase<D1> &     _bw1,
                                              const MatrixBase<D1> &     _c2,
                                              const MatrixBase<D1> &     _cd2,
                                              const MatrixBase<D1> &     _Lc2,
                                              const QuaternionBase<D3> & _q2,
                                              MatrixBase<D4> &           _res) const
{
    /*  Help for the Imu residual function
     *
     *  Notations:
     *    D_* : a motion in the Delta manifold                           -- implemented as 10-vectors with [Dp, Dq, Dv]
     *    T_* : a motion difference in the Tangent space to the manifold -- implemented as  9-vectors with [Dp, Do, Dv]
     *    b*  : a bias
     *    J*  : a Jacobian matrix
     *
     *  We use the following functions from imu_tools.h:
     *    D  = betweenStates(x1,x2,dt) : obtain a Delta from two states separated dt=t2-t1
     *    D2 = plus(D1,T)              : plus operator,  D2 = D1 (+) T
     *    T  = diff(D1,D2)             : minus operator, T  = D2 (-) D1
     *
     *  Two methods are possible (select with #define below this note) :
     *
     *  Computations common to methods 1 and 2:
     *    D_exp  = betweenStates(x1,x2,dt)   // Predict delta from states
     *    T_step = J_preint * (b - b_preint) // compute the delta correction step due to bias change
     *
     *  Method 1:
     *    D_corr = plus(D_preint, T_step)    // correct the pre-integrated delta with correction step due to bias change
     *    T_err  = diff(D_exp, D_corr)       // compare against expected delta
     *    res    = W.sqrt * T_err
     *
     *   results in:
     *    res    = W.sqrt * ( diff ( D_exp, plus(D_preint, J_preint * (b - b_preint) ) ) )
     *
     *  Method 2:
     *    T_diff = diff(D_preint, D_exp)     // compare pre-integrated against expected delta
     *    T_err  = T_diff - T_step           // the difference should match the correction step due to bias change
     *    res    = W.sqrt * err
     *
     *   results in :
     *    res    = W.sqrt * ( ( diff ( D_preint , D_exp ) ) - J_preint * (b - b_preint) )
     *
     * NOTE: See optimization report at the end of this file for comparisons of both methods.
     */

    //needed typedefs
    typedef typename D1::Scalar T;

    EIGEN_STATIC_ASSERT_VECTOR_SPECIFIC_SIZE(D4, 12);

    // 1. Expected delta from states
    Matrix<T, 3, 1 >   dc_exp;
    Matrix<T, 3, 1 >   dcd_exp;
    Matrix<T, 3, 1 >   dLc_exp;
    Quaternion<T>      dq_exp;

    bodydynamics::betweenStates(_c1, _cd1, _Lc1, _q1, _c2, _cd2, _Lc2, _q2, dt_, dc_exp, dcd_exp, dLc_exp, dq_exp);

    // 2. Corrected integrated delta: delta_corr = delta_preint (+) J_bias * (bias_current - bias_preint)

    // 2.a. Compute the delta step in tangent space:   step = J_bias * (bias - bias_preint)
    Matrix<T, 3, 1> dc_step  =                                          J_dc_wb_ * (_bw1 - gyro_bias_preint_);
    Matrix<T, 3, 1> dcd_step =                                         J_dcd_wb_ * (_bw1 - gyro_bias_preint_);
    Matrix<T, 3, 1> dLc_step = J_dLc_pb_ * (_bp1 - pbc_bias_preint_) + J_dLc_wb_ * (_bw1 - gyro_bias_preint_);
    Matrix<T, 3, 1> do_step  =                                          J_dq_wb_ * (_bw1 - gyro_bias_preint_);

    // 2.b. Add the correction step to the preintegrated delta:    delta_cor = delta_preint (+) step
    Matrix<T,3,1> dc_correct;
    Matrix<T,3,1> dcd_correct;
    Matrix<T,3,1> dLc_correct;
    Quaternion<T> dq_correct;

    bodydynamics::plus(dc_preint_.cast<T>(), dcd_preint_.cast<T>(), dLc_preint_.cast<T>(), dq_preint_.cast<T>(),
              dc_step, dcd_step, dLc_step, do_step,
              dc_correct, dcd_correct, dLc_correct, dq_correct);

    // 3. Delta error in minimal form: D_err = diff(D_exp , D_corr)
    // Note the Dt here is zero because it's the delta-time between the same time stamps!
    Matrix<T, 12, 1> d_error;
    Map<Matrix<T, 3, 1> >   dc_error (d_error.data()    );
    Map<Matrix<T, 3, 1> >   dcd_error(d_error.data() + 3);
    Map<Matrix<T, 3, 1> >   dLc_error(d_error.data() + 6);
    Map<Matrix<T, 3, 1> >   do_error (d_error.data() + 9);


    bodydynamics::diff(dc_exp, dcd_exp, dLc_exp, dq_exp,
              dc_correct, dcd_correct, dLc_correct, dq_correct, 
              dc_error, dcd_error, dLc_error, do_error);

    _res = getMeasurementSquareRootInformationUpper() * d_error;

    return true;
}

// Matrix<double,12,1> FactorForceTorquePreint::residual() const
// {
//     Matrix<double,12,1> res;


//     FrameBasePtr frm_prev = getFrameOther();
//     FrameBasePtr frm_curr = getFeature()->getFrame();

//     CaptureBaseWPtrList cap_lst = getCaptureOtherList();  // !! not retrieving the rigt captures...
//     auto cap_ikin_other = cap_lst.front().lock(); 
//     auto cap_gyro_other = cap_lst.back().lock();

//     Map<const Matrix<double,3,1> > c1( frm_prev->getStateBlock('C')->getState().data());
//     Map<const Matrix<double,3,1> > cd1(frm_prev->getStateBlock('D')->getState().data());
//     Map<const Matrix<double,3,1> > Lc1(frm_prev->getStateBlock('L')->getState().data());
//     Map<const Quaternion<double> > q1( frm_prev->getStateBlock('O')->getState().data());
//     Map<const Matrix<double,3,1> > bp1(cap_ikin_other->getStateBlock('I')->getState().data());
//     Map<const Matrix<double,3,1> > bw1(cap_gyro_other->getStateBlock('I')->getState().data() + 3);  // bw1 = bimu = [ba, bw]
//     Map<const Matrix<double,3,1> > c2 (frm_curr->getStateBlock('C')->getState().data());
//     Map<const Matrix<double,3,1> > cd2(frm_curr->getStateBlock('D')->getState().data());
//     Map<const Matrix<double,3,1> > Lc2(frm_curr->getStateBlock('L')->getState().data());
//     Map<const Quaternion<double> > q2 (frm_curr->getStateBlock('O')->getState().data());

//     residual(c1, cd1, Lc1, q1, bp1, bw1, c2, cd2, Lc2, q2, res);

//     return res;
// }

// double FactorForceTorquePreint::cost() const
// {
//     Matrix<double,12,1> toto = residual();
// }

} // namespace wolf

#endif
