//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#ifndef PROCESSOR_FORCE_TORQUE_PREINT_H
#define PROCESSOR_FORCE_TORQUE_PREINT_H

// Wolf core
#include <core/processor/processor_motion.h>

namespace wolf {
WOLF_STRUCT_PTR_TYPEDEFS(ParamsProcessorForceTorquePreint);

struct ParamsProcessorForceTorquePreint : public ParamsProcessorMotion
{
        std::string sensor_ikin_name;
        std::string sensor_angvel_name;
        int nbc;  // Number of contacts
        int dimc; // Dimension of the contact: 3D == point feet = 3D force, 6D = humanoid = wrench

        ParamsProcessorForceTorquePreint() = default;
        ParamsProcessorForceTorquePreint(std::string _unique_name, const ParamsServer& _server):
            ParamsProcessorMotion(_unique_name, _server)
        {
            sensor_ikin_name   = _server.getParam<std::string>(_unique_name + "/sensor_ikin_name");
            sensor_angvel_name = _server.getParam<std::string>(_unique_name + "/sensor_angvel_name");
            nbc                = _server.getParam<int>(_unique_name + "/nbc_");
            dimc               = _server.getParam<int>(_unique_name + "/dimc_");
        }
        ~ParamsProcessorForceTorquePreint() override = default;
        std::string print() const override
        {
            return ParamsProcessorMotion::print()                    + "\n"
                    + "sensor_ikin_name: "           + sensor_ikin_name     + "\n"
                    + "sensor_angvel_name: "         + sensor_angvel_name   + "\n"
                    + "nbc_: "                       + std::to_string(nbc)                  + "\n"
                    + "dimc_: "                      + std::to_string(dimc)                 + "\n";

        }
};

WOLF_PTR_TYPEDEFS(ProcessorForceTorquePreint);
    
//class
class ProcessorForceTorquePreint : public ProcessorMotion{
    public:
        ProcessorForceTorquePreint(ParamsProcessorForceTorquePreintPtr _params_motion_force_torque_preint);
        ~ProcessorForceTorquePreint() override;
        void configure(SensorBasePtr _sensor) override;

        WOLF_PROCESSOR_CREATE(ProcessorForceTorquePreint, ParamsProcessorForceTorquePreint);

    protected:
        void computeCurrentDelta(const Eigen::VectorXd& _data,
                                         const Eigen::MatrixXd& _data_cov,
                                         const Eigen::VectorXd& _calib,
                                         const double _dt,
                                         Eigen::VectorXd& _delta,
                                         Eigen::MatrixXd& _delta_cov,
                                         Eigen::MatrixXd& _jacobian_calib) const override;
        void deltaPlusDelta(const Eigen::VectorXd& _delta_preint,
                                    const Eigen::VectorXd& _delta,
                                    const double _dt,
                                    Eigen::VectorXd& _delta_preint_plus_delta) const override;
        void deltaPlusDelta(const Eigen::VectorXd& _delta_preint,
                                    const Eigen::VectorXd& _delta,
                                    const double _dt,
                                    Eigen::VectorXd& _delta_preint_plus_delta,
                                    Eigen::MatrixXd& _jacobian_delta_preint,
                                    Eigen::MatrixXd& _jacobian_delta) const override;
        void statePlusDelta(const VectorComposite& _x,
                                const Eigen::VectorXd& _delta,
                                const double _dt,
                                VectorComposite& _x_plus_delta ) const override;
        Eigen::VectorXd deltaZero() const override;
        Eigen::VectorXd correctDelta(const Eigen::VectorXd& delta_preint,
                                             const Eigen::VectorXd& delta_step) const override;
        VectorXd getCalibration (const CaptureBaseConstPtr _capture = nullptr) const override;
        void setCalibration(const CaptureBasePtr _capture, const VectorXd& _calibration) override;

        bool voteForKeyFrame() const override;
        CaptureMotionPtr emplaceCapture(const FrameBasePtr& _frame_own,
                                                const SensorBasePtr& _sensor,
                                                const TimeStamp& _ts,
                                                const VectorXd& _data,
                                                const MatrixXd& _data_cov,
                                                const VectorXd& _calib,
                                                const VectorXd& _calib_preint,
                                                const CaptureBasePtr& _capture_origin) override;
        FeatureBasePtr emplaceFeature(CaptureMotionPtr _capture_motion) override;
        FactorBasePtr emplaceFactor(FeatureBasePtr _feature_motion,
                                            CaptureBasePtr _capture_origin) override;

    private:
        ParamsProcessorForceTorquePreintPtr params_motion_force_torque_preint_;
        SensorBasePtr sensor_ikin_;
        SensorBasePtr sensor_angvel_;
        int nbc_;
        int dimc_;
};

}

/////////////////////////////////////////////////////////
// IMPLEMENTATION. Put your implementation includes here
/////////////////////////////////////////////////////////

namespace wolf{

inline void ProcessorForceTorquePreint::configure(SensorBasePtr _sensor)
{
    sensor_ikin_   = _sensor->getProblem()->findSensor(params_motion_force_torque_preint_->sensor_ikin_name);
    sensor_angvel_ = _sensor->getProblem()->findSensor(params_motion_force_torque_preint_->sensor_angvel_name);
};

inline Eigen::VectorXd ProcessorForceTorquePreint::deltaZero() const
{
    return (Eigen::VectorXd(13) << 0,0,0, 0,0,0,  0,0,0,  0,0,0,1 ).finished(); // com, com vel, ang momentum, orientation
}

} // namespace wolf

#endif // PROCESSOR_FORCE_TORQUE_PREINT_H
