//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/*
 * processor_preintegrator_force_torque_inertial.cpp
 *
 *  Created on: Aug 19, 2021
 *      Author: jsola
 */

// bodydynamics
#include "bodydynamics/processor/processor_force_torque_inertial_preint.h"
#include "bodydynamics/factor/factor_force_torque_inertial.h"
#include "bodydynamics/math/force_torque_inertial_delta_tools.h"

// wolf
#include <core/state_block/state_composite.h>
#include <core/capture/capture_motion.h>

namespace wolf
{

ProcessorForceTorqueInertialPreint::ProcessorForceTorqueInertialPreint(
    ParamsProcessorForceTorqueInertialPreintPtr _params_force_torque_inertial_preint)
    : ProcessorMotion("ProcessorForceTorqueInertialPreint",
                      "POVL",
                      3,   // dim
                      13,  // state size [p, q, v, L]
                      19,  // delta size [pi, vi, pd, vd, L, q]
                      18,  // delta cov size
                      12,  // data size [a, w, f, t]
                      6,   // calib size [ab, wb]
                      _params_force_torque_inertial_preint),
      params_force_torque_inertial_preint_(_params_force_torque_inertial_preint)
{
    // TODO Auto-generated constructor stub
}

ProcessorForceTorqueInertialPreint::~ProcessorForceTorqueInertialPreint()
{
    // TODO Auto-generated destructor stub
}

CaptureMotionPtr ProcessorForceTorqueInertialPreint::emplaceCapture(const FrameBasePtr&   _frame_own,
                                                                    const SensorBasePtr&  _sensor,
                                                                    const TimeStamp&      _ts,
                                                                    const VectorXd&       _data,
                                                                    const MatrixXd&       _data_cov,
                                                                    const VectorXd&       _calib,
                                                                    const VectorXd&       _calib_preint,
                                                                    const CaptureBasePtr& _capture_origin_ptr)
{
    CaptureMotionPtr capture = CaptureBase::emplace<CaptureMotion>(_frame_own, "CaptureForceTorqueInertial", _ts,
                                                                   _sensor, _data, _data_cov, _capture_origin_ptr);
    setCalibration(capture, _calib);
    capture->setCalibrationPreint(_calib_preint);
    return capture;
}

FeatureBasePtr ProcessorForceTorqueInertialPreint::emplaceFeature(CaptureMotionPtr _capture_own)
{
    FeatureBasePtr returnValue;
    return returnValue;
}

FactorBasePtr ProcessorForceTorqueInertialPreint::emplaceFactor(FeatureBasePtr _feature_motion,
                                                                CaptureBasePtr _capture_origin)
{
    FactorBasePtr returnValue;
    return returnValue;
}

void ProcessorForceTorqueInertialPreint::setCalibration(const CaptureBasePtr _capture, const VectorXd& _calibration) 
{
    _capture->getStateBlock('I')->setState(_calibration); // Set IMU bias
}

void ProcessorForceTorqueInertialPreint::configure(SensorBasePtr _sensor)
{
    sensor_fti_ = std::static_pointer_cast<SensorForceTorqueInertial>(_sensor);
}

void ProcessorForceTorqueInertialPreint::computeCurrentDelta(const Eigen::VectorXd& _data,
                                                             const Eigen::MatrixXd& _data_cov,
                                                             const Eigen::VectorXd& _calib,
                                                             const double           _dt,
                                                             Eigen::VectorXd&       _delta,
                                                             Eigen::MatrixXd&       _delta_cov,
                                                             Eigen::MatrixXd&       _jacobian_calib) const
{
    // compute tangent by removing IMU bias
    Matrix<double, 12, 1> tangent = _data;
    tangent.head<6>() -= _calib;            // J_tangent_data  = Identity(12,12)
    Matrix<double, 12, 6> J_tangent_calib;  // J_tangent_calib = [-Identity(6,6) ; Zero(6,6)]
    J_tangent_calib.topRows<6>()    = -Matrix6d::Identity();
    J_tangent_calib.bottomRows<6>() = Matrix6d::Zero();

    // go from tangent to delta. We need to dynamic model for this
    const Matrix<double, 7, 1>& model = sensor_fti_->getModel();
    Matrix<double, 18, 12>      J_delta_tangent;
    Matrix<double, 18, 7>       J_delta_model;
    fti::tangent2delta(tangent, model, _dt, _delta, J_delta_tangent, J_delta_model);

    // Compute cov and compose jacobians
    const auto& J_delta_data = J_delta_tangent;  //  * J_tangent_data, which is the Identity;
    _delta_cov               = J_delta_data * _data_cov * J_delta_data.transpose();
    _jacobian_calib          = J_delta_tangent * J_tangent_calib;
}

void ProcessorForceTorqueInertialPreint::deltaPlusDelta(const Eigen::VectorXd& _delta1,
                                                        const Eigen::VectorXd& _delta2,
                                                        const double           _dt2,
                                                        Eigen::VectorXd&       _delta1_plus_delta2) const
{
    bodydynamics::fti::compose(_delta1, _delta2, _dt2, _delta1_plus_delta2);
}

void ProcessorForceTorqueInertialPreint::deltaPlusDelta(const Eigen::VectorXd& _delta1,
                                                        const Eigen::VectorXd& _delta2,
                                                        const double           _dt2,
                                                        Eigen::VectorXd&       _delta1_plus_delta2,
                                                        Eigen::MatrixXd&       _jacobian1,
                                                        Eigen::MatrixXd&       _jacobian2) const
{
    bodydynamics::fti::compose(_delta1, _delta2, _dt2, _delta1_plus_delta2, _jacobian1, _jacobian2);
}

void ProcessorForceTorqueInertialPreint::statePlusDelta(const VectorComposite& _x,
                                                        const Eigen::VectorXd& _delta,
                                                        const double           _dt,
                                                        VectorComposite&       _x_plus_delta) const
{
    auto     x            = _x.vector("PVLO");
    VectorXd x_plus_delta = bodydynamics::fti::composeOverState(x, _delta, _dt);
    _x_plus_delta         = VectorComposite(x_plus_delta, "PVLO", {3, 3, 3, 4});
}

Eigen::VectorXd ProcessorForceTorqueInertialPreint::correctDelta(const Eigen::VectorXd& _delta_preint,
                                                                 const Eigen::VectorXd& _delta_step) const
{
    return fti::plus(_delta_preint, _delta_step);
}

VectorXd ProcessorForceTorqueInertialPreint::getCalibration(const CaptureBaseConstPtr _capture) const
{
    if (_capture)
        return _capture->getStateBlock('I')->getState();  // IMU bias
    else
        return getSensor()->getStateBlock('I')->getState();  // IMU bias
}
} /* namespace wolf */

#include <core/processor/factory_processor.h>
namespace wolf
{
WOLF_REGISTER_PROCESSOR(ProcessorForceTorqueInertialPreint);
WOLF_REGISTER_PROCESSOR_AUTO(ProcessorForceTorqueInertialPreint);
}  // namespace wolf