//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
/**
 * \file gtest_feature_inertial_kinematics.cpp
 *
 *  Created on: March 9, 2020
 *      \author: mfourmy
 */


#include "bodydynamics/feature/feature_inertial_kinematics.h"

#include <core/utils/utils_gtest.h>
#include <core/utils/logging.h>

#include <Eigen/Dense>

using namespace wolf;
using namespace Eigen;


TEST(FeatureInertialKinematics, jacobian_covariance_and_constructor)
{   
    Vector3d pBC_meas = Vector3d::Random(); 
    Vector3d vBC_meas = Vector3d::Random();
    Vector3d w_meas =   Vector3d::Random(); 
    
    Eigen::Matrix3d Qp = pow(1e-2, 2)*Eigen::Matrix3d::Identity();
    Eigen::Matrix3d Qv = pow(2e-2, 2)*Eigen::Matrix3d::Identity();
    Eigen::Matrix3d Qw = pow(3e-2, 2)*Eigen::Matrix3d::Identity();
    
    // momentum parameters
    Matrix3d Iq = Matrix3d::Identity(); 
    Vector3d Lq = Vector3d::Random(); 
    Vector9d meas_ikin; meas_ikin << pBC_meas, vBC_meas, w_meas;
    
    // direct computation of error covariance (analytical formula of first order propagation)
    Eigen::Matrix9d Q_ikin_err = computeIKinCov(Qp, Qv, Qw, pBC_meas, w_meas, Iq);

    // covariance from the first order propagation (using jacobian analytical formula)
    Matrix9d J_err_noise = computeIKinJac(pBC_meas, w_meas, Iq);
    Eigen::Matrix9d Q_meas = Eigen::Matrix9d::Zero(); 
    Q_meas.block<3,3>(0,0) = Qp;
    Q_meas.block<3,3>(3,3) = Qv;
    Q_meas.block<3,3>(6,6) = Qw;
    Eigen::Matrix9d Q_ikin_err_from_Jc = J_err_noise * Q_meas * J_err_noise.transpose();
    FeatureInertialKinematicsPtr feat_ptr = std::make_shared<FeatureInertialKinematics>(meas_ikin, Q_ikin_err, Iq, Lq);

    // Test if analytic covariance matches with the explicit jacobian propagation formula
    ASSERT_MATRIX_APPROX(Q_ikin_err, Q_ikin_err_from_Jc, 1e-3);

    // Simple constructor tests
    ASSERT_MATRIX_APPROX(Iq, feat_ptr->getBIq(), 1e-3);
    ASSERT_MATRIX_APPROX(Lq, feat_ptr->getBLcm(), 1e-3);
    Iq = 2*Matrix3d::Identity(); 
    Lq = Vector3d::Random(); 
    feat_ptr->setBIq(Iq);
    feat_ptr->setBLcm(Lq);
    ASSERT_MATRIX_APPROX(Iq, feat_ptr->getBIq(), 1e-3);
    ASSERT_MATRIX_APPROX(Lq, feat_ptr->getBLcm(), 1e-3);

}

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}

